from __future__ import absolute_import, division, print_function

import pytest

import yakonfig
from yakonfig import ConfigurationError, ProgrammerError
from yakonfig.factory import AutoFactory, AutoConfigured


def test_no_tuple_unpacking():
    def fun((a, b)): pass
    with pytest.raises(ProgrammerError):
        AutoConfigured(fun)


def test_no_var_args():
    def fun(*args): pass
    with pytest.raises(ProgrammerError):
        AutoConfigured(fun)


def test_no_var_kw_args():
    def fun(**kwargs): pass
    with pytest.raises(ProgrammerError):
        AutoConfigured(fun)


def test_bad_class():
    class OldStyle:
        pass
    with pytest.raises(ProgrammerError):
        AutoConfigured(OldStyle)


def configurable_defaults(a=1, b=2, c=3):
    return {'a': a, 'b': b, 'c': c}


def configurable_services(abc, xyz):
    return {'abc': abc, 'xyz': xyz}


def configurable_both(abc, xyz, a=1, b=2, c=3):
    return dict(configurable_services(abc, xyz),
                **configurable_defaults(a=a, b=b, c=c))


class configurable_class(object):
    def __init__(self, k='v'):
        self.k = k


class ConfigurableAltName(object):
    config_name = 'configurable_alt_name'
    def __init__(self, key='value'):
        self.key = key


def test_discover_defaults():
    conf = AutoConfigured(configurable_defaults)
    assert conf._discovered == {
        'name': 'configurable_defaults',
        'required': [],
        'defaults': {'a': 1, 'b': 2, 'c': 3},
    }


def test_discover_services():
    conf = AutoConfigured(configurable_services)
    assert conf._discovered == {
        'name': 'configurable_services',
        'required': ['abc', 'xyz'],
        'defaults': {},
    }


def test_discover_both():
    conf = AutoConfigured(configurable_both)
    assert conf._discovered == {
        'name': 'configurable_both',
        'required': ['abc', 'xyz'],
        'defaults': {'a': 1, 'b': 2, 'c': 3},
    }


def test_discover_class():
    conf = AutoConfigured(configurable_class)
    assert conf._discovered == {
        'name': 'configurable_class',
        'required': [],
        'defaults': {'k': 'v'},
    }

def test_discover_class_alt_name():
    conf = AutoConfigured(ConfigurableAltName)
    assert conf._discovered == {
        'name': 'configurable_alt_name',
        'required': [],
        'defaults': {'key': 'value'},
    }


def create_factory(configurables):
    class SimpleAutoFactory (AutoFactory):
        config_name = 'SimpleAutoFactory'
        @property
        def auto_config(self):
            return configurables
    return SimpleAutoFactory()


def test_factory_defaults():
    factory = create_factory([configurable_defaults])
    config = {'SimpleAutoFactory': {'configurable_defaults': {'b': 42}}}
    with yakonfig.defaulted_config([factory], config=config):
        instantiated = factory.create(configurable_defaults)
        assert instantiated == configurable_defaults(b=42)


def test_factory_defaults_override():
    factory = create_factory([configurable_defaults])
    config = {'SimpleAutoFactory': {'configurable_defaults': {'b': 42}}}
    with yakonfig.defaulted_config([factory], config=config):
        instantiated = factory.create(configurable_defaults, b=43)
        assert instantiated == configurable_defaults(b=43)


def test_factory_without_config():
    factory = create_factory([configurable_defaults])
    with pytest.raises(yakonfig.ProgrammerError):
        factory.create(configurable_defaults)


def test_factory_explicit_config():
    factory = create_factory([configurable_defaults])
    factory.config = {'configurable_defaults': {'b': 42}}
    instantiated = factory.create(configurable_defaults)
    assert instantiated['a'] == 1
    assert instantiated['b'] == 42
    assert instantiated['c'] == 3


def test_factory_param_config():
    factory = create_factory([configurable_defaults])
    instantiated = factory.create(configurable_defaults, config={'b': 42})
    assert instantiated['a'] == 1
    assert instantiated['b'] == 42
    assert instantiated['c'] == 3


def test_factory_services():
    factory = create_factory([configurable_services])
    factory.abc = 'abc'
    factory.xyz = 'xyz'
    with yakonfig.defaulted_config([factory]):
        instantiated = factory.create(configurable_services)
        assert instantiated == configurable_services('abc', 'xyz')


def test_factory_services_override():
    factory = create_factory([configurable_services])
    factory.abc = 'abc'
    factory.xyz = 'xyz'
    with yakonfig.defaulted_config([factory]):
        instantiated = factory.create(configurable_services, xyz='foo')
        assert instantiated == configurable_services('abc', 'foo')


def test_factory_defaults_and_services():
    factory = create_factory([configurable_both])
    factory.abc = 'abc'
    factory.xyz = 'xyz'
    config = {'SimpleAutoFactory': {'configurable_both': {'c': 42}}}
    with yakonfig.defaulted_config([factory], config=config):
        instantiated = factory.create(configurable_both)
        assert instantiated == configurable_both('abc', 'xyz', c=42)


def test_factory_missing_service():
    factory = create_factory([configurable_services])
    # Not adding any services to `factory`...
    with yakonfig.defaulted_config([factory]):
        with pytest.raises(ProgrammerError):
            factory.create(configurable_services)


def test_factory_service_config_conflict():
    factory = create_factory([configurable_services])
    factory.abc = 'abc'
    factory.xyz = 'xyz'
    config = {'SimpleAutoFactory': {'configurable_services': {'abc': 'abc'}}}
    with pytest.raises(ConfigurationError):
        yakonfig.set_default_config([factory], config=config)


def test_factory_extra_config():
    factory = create_factory([configurable_defaults])
    config = {'SimpleAutoFactory': {'configurable_defaults': {'ZZZ': 42}}}
    with pytest.raises(ConfigurationError):
        yakonfig.set_default_config([factory], config=config)


def test_factory_class():
    factory = create_factory([configurable_class])
    with yakonfig.defaulted_config([factory], config={}) as config:
        assert 'SimpleAutoFactory' in config
        assert 'configurable_class' in config['SimpleAutoFactory']
        assert config['SimpleAutoFactory']['configurable_class']['k'] == 'v'
        instantiated = factory.create(configurable_class)
        assert instantiated.k == 'v'
        instantiated = factory.create(configurable_class, k='k')
        assert instantiated.k == 'k'


def test_factory_class_by_name():
    factory = create_factory([configurable_class])
    with yakonfig.defaulted_config([factory], config={}):
        instantiated = factory.create('configurable_class')
        assert isinstance(instantiated, configurable_class)
        assert instantiated.k == 'v'


def test_factory_class_with_config_name():
    factory = create_factory([ConfigurableAltName])
    with yakonfig.defaulted_config([factory], config={}) as config:
        assert 'SimpleAutoFactory' in config
        assert 'configurable_alt_name' in config['SimpleAutoFactory']
        assert (config['SimpleAutoFactory']['configurable_alt_name']['key'] ==
                'value')
        instantiated = factory.create(ConfigurableAltName)
        assert instantiated.key == 'value'
        instantiated = factory.create(ConfigurableAltName, key='key')
        assert instantiated.key == 'key'


class FactoryWithProperty(AutoFactory):
    config_name = 'factory_with_property'
    auto_config = [configurable_services]
    @property
    def abc(self):
        return 'foo'
    @property
    def xyz(self):
        return 'bar'


def test_factory_with_property():
    factory = FactoryWithProperty()
    with yakonfig.defaulted_config([factory]):
        instantiated = factory.create(configurable_services)
        assert instantiated['abc'] == 'foo'
        assert instantiated['xyz'] == 'bar'



"""Configuration merging and diffing.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

Purpose
=======

This module provides common functions to add one configuration block
to another (:func:`overlay_config`), and to do the reverse
(:func:`diff_config`).

Module Contents
===============
"""
from __future__ import absolute_import
import collections

def overlay_config(c0, c1):
    """Overlay one configuration over another.

    This overlays `c1` on top of `c0` as follows:

    * If either isn't a dictionary, returns `c1`.
    * Any key in `c0` not present in `c1` is present in the result with
      its original value.
    * Any key in `c1` with value :const:`None` is not present in the result.
    * Any key in `c1` not present in `c0` and not :const:`None` is present in
      the result with its new value.
    * Any key in both `c1` and `c0` with a non-:const:`None` value is
      recursively overlaid.

    >>> overlay_config({'a': 'b'}, {'a': 'c'})
    {'a': 'c'}
    >>> overlay_config({'a': 'b'}, {'c': 'd'})
    {'a': 'b', 'c': 'd'}
    >>> overlay_config({'a': {'b': 'c'}},
    ...                {'a': {'b': 'd', 'e': 'f'}})
    {'a': {'b': 'd', 'e': 'f'}}
    >>> overlay_config({'a': 'b', 'c': 'd'}, {'a': None})
    {'c': 'd'}

    :param dict c0: original configuration
    :param dict c1: overlay configuration
    :return: new overlaid configuration
    :returntype dict:

    """
    if not isinstance(c0, collections.Mapping): return c1
    if not isinstance(c1, collections.Mapping): return c1
    result = dict()
    for k in c0.iterkeys():
        if k not in c1:
            result[k] = c0[k]
    for k,v in c1.iteritems():
        if v is not None:
            if k in c0:
                v = overlay_config(c0[k], v)
            result[k] = v
    return result

def diff_config(c0, c1):
    """Find the differences between two configurations.

    This finds a delta configuration from `c0` to `c1`, such that
    calling :func:`overlay_config` with `c0` and the result of this
    function yields `c1`.  This works as follows:

    * If both are identical (of any type), returns an empty dictionary.
    * If either isn't a dictionary, returns `c1`.
    * Any key in `c1` not present in `c0` is included in the output
      with its value from `c1`.
    * Any key in `c0` not present in `c1` is included in the output
      with value :const:`None`.
    * Any keys present in both dictionaries are recursively merged.

    >>> diff_config({'a': 'b'}, {})
    {'a': None}
    >>> diff_config({'a': 'b'}, {'a': 'b', 'c': 'd'})
    {'c': 'd'}

    :param dict c0: original configuration
    :param dict c1: new configuration
    :return: overlay configuration
    :returntype dict:

    """
    if not isinstance(c0, collections.Mapping):
        if c0 == c1:
            return {}
        return c1
    if not isinstance(c1, collections.Mapping): return c1
    result = dict()
    for k in c0.iterkeys():
        if k not in c1:
            result[k] = None
    for k,v in c1.iteritems():
        if k in c0:
            merged = diff_config(c0[k], v)
            if merged != {}: result[k] = merged
        else:
            result[k] = v
    return result



'''Base class for :mod:`argparse`-based interactive :mod:`cmd` sessions.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

.. autoclass:: ArgParseCmd

'''
from __future__ import absolute_import
import argparse
from cmd import Cmd
import logging
import pdb
import shlex

class ArgParseCmd(Cmd):
    '''Base class for :mod:`argparse`-based interactive :mod:`cmd` sessions.

    This extends :class:`cmd.Cmd` to allow individual commands to
    configure an :class:`argparse.ArgumentParser` object.  If a command
    has a ``args_command()`` method in parallel with its ``do_command()``
    method, then:

    * ``args_command()`` is called with an :class:`~argparse.ArgumentParser`
      object as a parameter and may add arguments to it

    * ``do_command()`` takes a single argument, which is the parsed
      :class:`argparse.Namespace`

    For instance, a typical program could look like:

    .. code-block:: python

        class AProgram(ArgParseCmd):
            def __init__(self):
                ArgParseCmd.__init__(self)
                self.prompt = 'aprogram> '

            def args_command(self, parser):
                parser.add_argument('--frobnicate', action='store_true',
                    help='frobnicate the frabulator')
            def do_command(self, args):
                """do work with the frabulator"""
                if args.frobnicate:
                    self.stdout.write('frobnicating\\n')
                else:
                    self.stdout.write('not frobnicating\\n')

    Running ``help command`` or ``command --help`` will print the
    argparse-generated help, using the main function's docstring as
    the short description.  Input lines are split with the
    :class:`shlex` module, and so complex strings can be quoted.  If
    there is no ``args_...()`` function, the complete line is passed
    to the ``do_...()`` function, just as in the base :class:`cmd.Cmd`
    class.  If your command takes no parameters, providing an empty
    ``args_...()`` function is good practice.

    .. note:: In Python 2.7, :class:`cmd.Cmd` is an old-style class,
              and so :func:`super` and other new-style class features
              will not work with this class.

    In addition to ``help``, this provides a standard ``quit`` command,
    and rebinds end-of-file to be equivalent to ``quit``.

    For standalone programs, the external shell has generally already
    parsed the command line into a sequence of arguments.
    :meth:`add_arguments` adds required arguments to a
    :class:`argparse.ArgumentParser`, and :meth:`main` either runs the
    main loop or the requested command.

    .. code-block:: python

        parser = argparse.ArgumentParser()
        app = AProgram()
        app.add_arguments(parser)
        args = yakonfig.parse_args(parser, [yakonfig])
        app.main(args)

    .. automethod:: add_arguments
    .. automethod:: main
    .. automethod:: runcmd

    '''
    def add_arguments(self, parser):
        '''Add generic command-line arguments to a top-level argparse parser.

        After running this, the results from ``argparse.parse_args()``
        can be passed to :meth:`main`.

        '''
        commands = set(name[3:] for name in dir(self) if name.startswith('do_'))
        parser.add_argument('action', help='action to run', nargs='?',
                            choices=list(commands))
        parser.add_argument('arguments', help='arguments specific to ACTION',
                            nargs=argparse.REMAINDER)

    def main(self, args):
        '''Run a single command, or else the main shell loop.

        `args` should be the :class:`argparse.Namespace` object after
        being set up via :meth:`add_arguments`.

        '''
        if args.action:
            self.runcmd(args.action, args.arguments)
        else:
            self.cmdloop()

    def runcmd(self, cmd, args):
        '''Run a single command from pre-parsed arguments.

        This is intended to be run from :meth:`main` or somewhere else
        "at the top level" of the program.  It may raise
        :exc:`exceptions.SystemExit` if an argument such as ``--help``
        that normally causes execution to stop is encountered.

        '''
        dof = getattr(self, 'do_' + cmd, None)
        if dof is None:
            return self.default(' '.join([cmd] + args))
        argf = getattr(self, 'args_' + cmd, None)
        if argf is not None:
            parser = argparse.ArgumentParser(
                prog=cmd,
                description=getattr(dof, '__doc__', None))
            argf(parser)
            argl = parser.parse_args(args)
        else:
            argl = ' '.join(args)
        return dof(argl)

    def parseline(self, line):
        cmd, arg, line = Cmd.parseline(self, line)
        if cmd and cmd.strip() != '':
            dof = getattr(self, 'do_' + cmd, None)
            argf = getattr(self, 'args_' + cmd, None)
        else:
            argf = None
        if argf:
            parser = argparse.ArgumentParser(
                prog=cmd,
                description=getattr(dof, '__doc__', None))
            argf(parser)
            try:
                arg = parser.parse_args(shlex.split(arg))
            except SystemExit, e:
                return '', '', ''
        return cmd, arg, line

    def emptyline(self):
        pass

    def args_help(self, parser):
        parser.add_argument('command', nargs='?',
                            help='print help on this command')
    def do_help(self, args):
        '''print help on a command'''
        if args.command:
            f = getattr(self, 'help_' + args.command, None)
            if f:
                f()
                return

            f = getattr(self, 'do_' + args.command, None)
            if not f:
                msg = self.nohelp % (args.command,)
                self.stdout.write('{}\n'.format(msg))
                return

            docstr = getattr(f, '__doc__', None)
            f = getattr(self, 'args_' + args.command, None)
            if f:
                parser = argparse.ArgumentParser(
                    prog=args.command,
                    description=docstr)
                f(parser)
                parser.print_help(file=self.stdout)
            else:
                if not docstr:
                    docstr = self.nohelp % (args.command,)
                self.stdout.write('{}\n'.format(docstr))
        else:
            Cmd.do_help(self, '')

    def precmd(self, line):
        if line == 'EOF': return 'quit'
        return line

    def args_quit(self, parser): pass
    def do_quit(self, line):
        '''exit the program'''
        return True



"""Tests of various paths through yakonfig.Configurable.

-----

This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.

"""

from __future__ import absolute_import

import argparse

import pytest

import yakonfig
import yakonfig.tests.configurable_module
import yakonfig.toplevel

class ConfigurableSubclass(yakonfig.Configurable):
    @property
    def config_name(self): return 'configurable'
    @property
    def default_config(self): return { 'type': 'object' }

class ConfigurableLike(object):
    config_name = 'configurable'
    default_config = { 'type': 'class' }
    runtime_keys = { 'key': 'key' }

class ConfigurableArgs(yakonfig.Configurable):
    @property
    def config_name(self): return 'config'
    @property
    def default_config(self): return {'k': 'k'}
    def add_arguments(self, parser):
        parser.add_argument('-k', '--key', metavar='VALUE')
    @property
    def runtime_keys(self): return {'key': 'k'}
    def check_config(self, config, name=''):
        if len(config['k']) != 1:
            raise yakonfig.ConfigurationError("{} 'k' wrong length"
                                              .format(name))

class ConfigurableBottom(object):
    config_name = 'bottom'
    default_config = { 'zzz': '-32768' }
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--zzz', '-z')
    runtime_keys = { 'zzz': 'zzz' }

class ConfigurableTop(object):
    config_name = 'top'
    default_config = { 'aaa': 'bbb' }
    sub_modules = [ConfigurableBottom]
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--aaa', '-a')
    runtime_keys = { 'aaa': 'aaa' }

class ConfigurableAlmostTop(object):
    config_name = 'top'
    default_config = { 'aaa': 'bbb' }
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--aaa', '-a')
    runtime_keys = { 'aaa': 'aaa' }
    @staticmethod
    def replace_config(config, name):
        return ConfigurableTop

class ConfigurableLikeTop(object):
    config_name = 'top'
    default_config = { 'aaa': 'bbb' }
    @staticmethod
    def add_arguments(parser):
        parser.add_argument('--aaa', '-a')
    runtime_keys = { 'aaa': 'aaa' }
    @staticmethod
    def replace_config(config, name):
        return yakonfig.NewSubModules(ConfigurableLikeTop, [ConfigurableBottom])

class Dependent(object):
    config_name = 'dependent'
    @staticmethod
    def check_config(config, name):
        yakonfig.check_toplevel_config(ConfigurableArgs(), name)

class Normalized(object):
    config_name = 'normalized'
    default_config = { 'k': 'value' }
    runtime_keys = { 'k': 'k' }
    @staticmethod
    def normalize_config(config):
        config['k'] = config['k'][0]

class Discovers(object):
    config_name = 'discovers'
    default_config = { 'a': 'one', 'c': 'three' }
    runtime_keys = { 'a': 'a', 'b': 'b', 'c': 'c' }
    @staticmethod
    def discover_config(config, name):
        if 'a' not in config:
            config['a'] = 'foo'
        if 'b' not in config:
            config['b'] = 'bar'

@pytest.fixture(params=['object', 'class', 'module'])
def configurable_type(request):
    return request.param

@pytest.fixture
def a_configurable(configurable_type):
    if configurable_type == 'object': return ConfigurableSubclass()
    if configurable_type == 'class': return ConfigurableLike
    if configurable_type == 'module': return yakonfig.tests.configurable_module
    raise RuntimeError(t)

@pytest.yield_fixture
def global_yakonfig(a_configurable):
    yakonfig.set_default_config([a_configurable])
    yield yakonfig.get_global_config()
    yakonfig.clear_global_config()

def test_assemble_default_config(a_configurable, configurable_type):
    c = yakonfig.toplevel.assemble_default_config([a_configurable])
    assert sorted(c.iterkeys()) == ['configurable']
    cc = c['configurable']
    assert sorted(cc.iterkeys()) == ['type']
    assert cc['type'] == configurable_type
    assert cc.get('key') is None

def test_assemble_minimal():
    class MinimallyConfigurable(object):
        config_name = 'minimal'
    c = yakonfig.toplevel.assemble_default_config([MinimallyConfigurable])
    assert sorted(c.iterkeys()) == ['minimal']
    assert c['minimal'] == {}

def test_assemble_broken():
    class NotReallyConfigurable(object):
        pass
    with pytest.raises(yakonfig.ProgrammerError):
        yakonfig.toplevel.assemble_default_config([NotReallyConfigurable])

def test_assemble_two():
    c = yakonfig.toplevel.assemble_default_config([ConfigurableArgs(),
                                                   ConfigurableSubclass()])
    assert sorted(c.iterkeys()) == ['config', 'configurable']
    assert c['config'] == {'k': 'k'}
    assert c['configurable'] == {'type': 'object'}

def test_duplicates():
    with pytest.raises(yakonfig.ProgrammerError):
        yakonfig.toplevel.assemble_default_config([ConfigurableLike,
                                                   ConfigurableSubclass()])

def test_config_type(global_yakonfig, configurable_type):
    c = global_yakonfig
    assert sorted(c.iterkeys()) == ['configurable']
    cc = c['configurable']
    assert sorted(cc.iterkeys()) == ['type']
    assert cc['type'] == configurable_type
    assert cc.get('key') is None

def test_fill_in():
    c = { 'configurable': {} }
    yakonfig.toplevel.fill_in_arguments(c, [ConfigurableLike],
                                        { 'key': 'value' })
    assert sorted(c.iterkeys()) == ['configurable']
    cc = c['configurable']
    assert sorted(cc.iterkeys()) == ['key']
    assert cc['key'] == 'value'

def test_fill_in_replaces():
    c = { 'configurable': { 'key': 'old' } }
    yakonfig.toplevel.fill_in_arguments(c, [ConfigurableLike],
                                        { 'key': 'new' })
    assert sorted(c.iterkeys()) == ['configurable']
    cc = c['configurable']
    assert sorted(cc.iterkeys()) == ['key']
    assert cc['key'] == 'new'

def test_fill_in_object():
    class Params(object):
        def __init__(self):
            self.key = 'value'
    c = { 'configurable': {} }
    yakonfig.toplevel.fill_in_arguments(c, [ConfigurableLike], Params())
    assert sorted(c.iterkeys()) == ['configurable']
    cc = c['configurable']
    assert sorted(cc.iterkeys()) == ['key']
    assert cc['key'] == 'value'    

def test_default_file(request):
    yakonfig.set_default_config(
        [ConfigurableArgs()],
        filename=str(request.fspath.dirpath('argconfig.yaml')))
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['config']
        cc = c['config']
        assert sorted(cc.iterkeys()) == ['k']
        assert cc['k'] == 'x'
    finally:
        yakonfig.clear_global_config()

def test_fill_in_kvp():
    yakonfig.set_default_config([ConfigurableLike], { 'key': 'value' })
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['configurable']
        cc = c['configurable']
        assert sorted(cc.iterkeys()) == ['key', 'type']
        assert cc['type'] == 'class'
        assert cc['key'] == 'value'
    finally:
        yakonfig.clear_global_config()

def test_prog_yaml():
    the_yaml = """
configurable:
    key: yaml
"""
    yakonfig.set_default_config([ConfigurableLike], yaml=the_yaml)
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['configurable']
        cc = c['configurable']
        assert sorted(cc.iterkeys()) == ['key', 'type']
        assert cc['type'] == 'class'
        assert cc['key'] == 'yaml'
    finally:
        yakonfig.clear_global_config()

def test_prog_yaml():
    # like from the command line, values via params overrides explicit yaml
    the_yaml = """
configurable:
    key: yaml
"""
    yakonfig.set_default_config([ConfigurableLike],
                                params={'key': 'value'},
                                yaml=the_yaml)
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['configurable']
        cc = c['configurable']
        assert sorted(cc.iterkeys()) == ['key', 'type']
        assert cc['type'] == 'class'
        assert cc['key'] == 'value'
    finally:
        yakonfig.clear_global_config()

def test_dont_validate():
    yakonfig.set_default_config([ConfigurableArgs()],
                                params={'key': 'longer than one char'},
                                validate=False)
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['config']
        cc = c['config']
        assert sorted(cc.iterkeys()) == ['k']
        assert cc['k'] == 'longer than one char'
    finally:
        yakonfig.clear_global_config()

def test_two_level():
    yakonfig.set_default_config([ConfigurableTop])
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['top']
        c = c['top']
        assert sorted(c.iterkeys()) == ['aaa', 'bottom']
        assert c['aaa'] == 'bbb'
        c = c['bottom']
        assert sorted(c.iterkeys()) == ['zzz']
        assert c['zzz'] == '-32768'
    finally:
        yakonfig.clear_global_config()

def test_two_level_args():
    yakonfig.set_default_config([ConfigurableTop],
                                { 'aaa': 'a', 'bbb': 'b', 'zzz': 'z' })
    try:
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['top']
        c = c['top']
        assert sorted(c.iterkeys()) == ['aaa', 'bottom']
        assert c['aaa'] == 'a'
        c = c['bottom']
        assert sorted(c.iterkeys()) == ['zzz']
        assert c['zzz'] == 'z'
    finally:
        yakonfig.clear_global_config()

def test_overlay():
    oc = yakonfig.toplevel.overlay_config
    # non-dictionary cases
    assert oc('foo', 'bar') == 'bar'
    assert oc({'k': 'v'}, 'bar') == 'bar'
    assert oc('foo', {'k': 'v'}) == {'k': 'v'}
    # add and remove
    assert oc({'a': 'a'}, {}) == {'a': 'a'}
    assert oc({'a': 'a'}, {'a': None}) == {}
    assert oc({'a': 'a'}, {'a': 'b'}) == {'a': 'b'}
    assert oc({'a': 'a'}, {'b': 'b'}) == {'a': 'a', 'b': 'b'}
    assert oc({'a': 'x', 'b': 'y'}, {}) == {'a': 'x', 'b': 'y'}
    assert oc({'a': 'x', 'b': 'y'}, {'a': None}) == {'b': 'y'}
    assert (oc({'a': 'x', 'b': 'y'}, {'b': 'foo', 'c': 'bar'}) ==
            {'a': 'x', 'b': 'foo', 'c': 'bar'})
    # subobjects
    assert (oc({'a': {'1': 'one'} }, {'a': {'2': 'two'} }) ==
            {'a': {'1': 'one', '2': 'two' }})
    assert (oc({'a': {'1': 'one'} }, {'a': {'1': None} }) ==
            {'a': {} })
    assert (oc({'a': {'1': 'one'} },
               {'a': {'2': 'two'}, 'b': {'3': 'three'} }) ==
            {'a': {'1': 'one', '2': 'two'}, 'b': {'3': 'three'} })

def test_yakonfig_default():
    yakonfig.set_default_config([yakonfig])
    try:
        c = yakonfig.get_global_config()
        assert 'yakonfig' in c
    finally:
        yakonfig.clear_global_config()

def test_yakonfig_cli():
    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser, [yakonfig], args=[])
    try:
        c = yakonfig.get_global_config()
        assert 'yakonfig' in c
    finally:
        yakonfig.clear_global_config()    

def test_cli_none():
    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser, [ConfigurableArgs()], args=[])
    try:
        c = yakonfig.get_global_config()
        assert 'config' in c
        assert 'k' in c['config']
        assert c['config']['k'] == 'k'
    finally:
        yakonfig.clear_global_config()

def test_cli_good():
    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser, [ConfigurableArgs()], args=['--key', 'v'])
    try:
        c = yakonfig.get_global_config()
        assert 'config' in c
        assert 'k' in c['config']
        assert c['config']['k'] == 'v'
    finally:
        yakonfig.clear_global_config()

def test_cli_bad():
    parser = argparse.ArgumentParser()
    with pytest.raises(SystemExit):
        yakonfig.parse_args(parser, [ConfigurableArgs()],
                            args=['--key', 'value'])
        yakonfig.clear_global_config()

def test_cli_file(request):
    yaml = str(request.fspath.dirpath('argconfig.yaml'))
    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser,
                        [yakonfig, ConfigurableArgs()],
                        args=['-c', yaml])
    try:
        c = yakonfig.get_global_config()
        assert 'config' in c
        assert 'k' in c['config']
        assert c['config']['k'] == 'x' # from the file
    finally:
        yakonfig.clear_global_config()

def test_cli_overlay(request):
    # config.k is in the default, *and* the config file, *and* the command line
    yaml = str(request.fspath.dirpath('argconfig.yaml'))
    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser,
                        [yakonfig, ConfigurableArgs()],
                        args=['-c', yaml, '-k', 'y'])
    try:
        c = yakonfig.get_global_config()
        assert 'config' in c
        assert 'k' in c['config']
        assert c['config']['k'] == 'y' # from the command line
    finally:
        yakonfig.clear_global_config()

def test_check_dependent():
    # give an invalid configuration for args; but it's not the "real"
    # application so it shouldn't be checked
    with yakonfig.defaulted_config([ConfigurableArgs(), ConfigurableLike],
                                   {'key': 'value'}) as config:
        assert sorted(config.iterkeys()) == ['config', 'configurable']
        with pytest.raises(yakonfig.ConfigurationError):
            yakonfig.check_toplevel_config(ConfigurableArgs(), 'test')
        with pytest.raises(yakonfig.ConfigurationError):
            yakonfig.check_toplevel_config(ConfigurableTop, 'test')

def test_check_toplevel():
    '''check_config_toplevel() should work in check_config() implementation'''
    with yakonfig.defaulted_config([ConfigurableArgs(), Dependent],
                                   {'key': 'k'}) as config:
        assert sorted(config.iterkeys()) == ['config', 'dependent']
        assert config['config']['k'] == 'k'
    with pytest.raises(yakonfig.ConfigurationError):
        with yakonfig.defaulted_config([ConfigurableArgs(), Dependent],
                                       {'key': 'key'}):
            pass

def test_check_toplevel_args():
    '''check_config_toplevel() should work in check_config() implementation'''
    parser = argparse.ArgumentParser()
    yakonfig.parse_args(parser, [ConfigurableArgs(), Dependent],
                        ['--key', 'k'])
    try:
        config = yakonfig.get_global_config()
        assert sorted(config.iterkeys()) == ['config', 'dependent']
        assert config['config']['k'] == 'k'
    finally:
        yakonfig.clear_global_config()
    parser = argparse.ArgumentParser()
    with pytest.raises(SystemExit):
        yakonfig.parse_args(parser, [ConfigurableArgs(), Dependent],
                            ['--key', 'key'])
        yakonfig.clear_global_config()

def test_proxy_two_level():
    with yakonfig.defaulted_config(
            [yakonfig.ProxyConfigurable(ConfigurableTop)],
            { 'aaa': 'a', 'bbb': 'b', 'zzz': 'z' }):
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['top']
        c = c['top']
        assert sorted(c.iterkeys()) == ['aaa', 'bottom']
        assert c['aaa'] == 'a'
        c = c['bottom']
        assert sorted(c.iterkeys()) == ['zzz']
        assert c['zzz'] == 'z'

def test_replaces_direct():
    with yakonfig.defaulted_config([ConfigurableAlmostTop]):
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['top']
        c = c['top']
        assert 'bottom' in c
        c = c['bottom']
        assert c['zzz'] == '-32768'

def test_replaces_proxy():
    with yakonfig.defaulted_config([ConfigurableLikeTop]):
        c = yakonfig.get_global_config()
        assert sorted(c.iterkeys()) == ['top']
        c = c['top']
        assert 'bottom' in c
        c = c['bottom']
        assert c['zzz'] == '-32768'

def test_normalize():
    with yakonfig.defaulted_config([Normalized]):
        assert yakonfig.get_global_config('normalized')['k'] == 'v'
    with yakonfig.defaulted_config([Normalized], { 'k': 'foo' }):
        assert yakonfig.get_global_config('normalized')['k'] == 'f'
    with yakonfig.defaulted_config([Normalized],
                                   yaml='''
normalized:
  k: zoom!
'''):
        assert yakonfig.get_global_config('normalized')['k'] == 'z'

def test_discovery():
    with yakonfig.defaulted_config([Discovers]):
        # discovered overwrites default
        assert yakonfig.get_global_config('discovers', 'a') == 'foo'
        # discovered provides missing value
        assert yakonfig.get_global_config('discovers', 'b') == 'bar'
        # undiscovered uses default value
        assert yakonfig.get_global_config('discovers', 'c') == 'three'
    with yakonfig.defaulted_config([Discovers], { 'a': 'alpha' }):
        # command-line value overrules discovery
        assert yakonfig.get_global_config('discovers', 'a') == 'alpha'
        assert yakonfig.get_global_config('discovers', 'b') == 'bar'
        assert yakonfig.get_global_config('discovers', 'c') == 'three'



from __future__ import absolute_import, division, print_function
import abc
import copy
import inspect

from yakonfig.configurable import Configurable
from yakonfig.exceptions import ConfigurationError, ProgrammerError


try:
    strtype = basestring
except NameError:
    strtype = str


class AutoFactory (Configurable):
    '''A factory for *discovering* configuration from functions, methods
    or classes.

    Clients that subclass :class:`AutoFactory` must implement the
    :attr:`auto_config` property, which should be an iterable of
    things to automatically a configuration from. Notably, subclasses
    should *not* implement :attr:`~yakonfig.Configurable.sub_modules`,
    as this class provides its own implementation of it using
    :attr:`auto_config`.

    This class hooks into the :mod:`yakonfig` configuration sequence
    to capture its part of the configuration at startup time.  If
    this class is used outside the standard sequence, then you must
    set :attr:`config` before calling :meth:`create`.

    Currently, this class does **not** support a hierarchical
    configuration: items in :attr:`auto_config` may not be
    :class:`~yakonfig.Configurable`, and :class:`AutoFactory` objects
    may not be nested.  This is likely to change in the future.

    '''
    __metaclass__ = abc.ABCMeta

    def __init__(self, config=None):
        '''Create a new object factory.

        If `config` is :const:`None` (the default), then the
        :attr:`config` property must be set before calling
        :meth:`create`.  Passing this factory object to
        :func:`yakonfig.parse_args` or a similar high-level
        :mod:`yakonfig` setup method will also accomplish this.

        :param dict config: local configuration dictionary (if available)
        
        '''
        super(AutoFactory, self).__init__()
        self._config = config

    @property
    def sub_modules(self):
        return [AutoConfigured(obj) for obj in self.auto_config]

    @abc.abstractproperty
    def auto_config(self):
        '''Must return a list of objects to automatically configure.

        This list is interpreted shallowly. That is, all configuration
        from each object is discovered through its name and parameter
        list only. Everything else is ignored.

        '''
        pass


    def check_config(self, config, prefix=''):
        for child in self.sub_modules:
            child.check_config(config)

    def normalize_config(self, config):
        '''Rewrite (and capture) the configuration of this object.'''
        self.config = config

    @property
    def config(self):
        '''Saved configuration for the factory and its sub-objects.'''
        if self._config is None:
            raise ProgrammerError(
                'Tried to access saved factory configuration before '
                'yakonfig configuration was run.')
        return self._config
    @config.setter
    def config(self, c):
        self._config = c
        self.new_config()

    def new_config(self):
        '''Hook called when the configuration changes.

        If factory implementations keep properties that are created
        from the configuration, this is a place to create or reset them.
        The base class implementation does nothing.

        '''
        pass

    def create(self, configurable, config=None, **kwargs):
        '''Create a sub-object of this factory.

        Instantiates the `configurable` object with the current saved
        :attr:`config`.  This essentially translates to
        ``configurable(**config)``, except services defined in the
        parent and requested by `configurable` (by setting the
        ``services`` attribute) are injected. If a service is not
        defined on this factory object, then a
        :exc:`yakonfig.ProgrammerError` is raised.

        If `config` is provided, it is a local configuration for
        `configurable`, and it overrides the saved local configuration
        (if any).  If not provided, then :attr:`config` must already be
        set, possibly by passing this object into the :mod:`yakonfig`
        top-level setup sequence.

        :param callable configurable: object to create
        :param dict config: local configuration for `configurable`
        :param kwargs: additional keyword parameters
        :return: ``configurable(**config)``

        '''
        # If we got passed a string, find the thing to make.
        if isinstance(configurable, str):
            candidates = [ac for ac in self.sub_modules
                          if ac.config_name == configurable]
            if len(candidates) == 0:
                raise KeyError(configurable)
            configurable = candidates[0]

        # Regenerate the configuration ifneedbe.
        if not isinstance(configurable, AutoConfigured):
            configurable = AutoConfigured(configurable)

        if config is None:
            config = self.config.get(configurable.config_name, {})
        # shallow-copy config and append kwargs to it
        config = dict(config, **kwargs)
        for other in getattr(configurable, 'services', []):
            # AutoConfigured.check_config() validates that this key
            # wasn't in the global config, so this must have come from
            # either our own config parameter, a keyword arg, or
            # the caller setting factory.config; trust those paths.
            if other not in config:
                try:
                    config[other] = getattr(self, other)
                except AttributeError:
                    raise ProgrammerError(
                        'Configured object "%s" expects a '
                        '"%s" object to be available (from its '
                        'parameter list), but "%s" does not '
                        'provide it.'
                        % (repr(configurable.obj), other, repr(self)))
        return configurable(**config)


class AutoConfigured (Configurable):
    '''
    This is an **unexported** wrapper class that provides an
    implementation that satisfies :class:`yakonfig.Configurable`
    for objects that can have their configuration automatically
    discovered.
    '''
    def __init__(self, obj):
        self.obj = obj
        self._discovered = self._discover_config()
        self._config_name = self._discovered['name']
        self._services = self._discovered['required']
        self._default_config = self._discovered['defaults']

    def __call__(self, *args, **kwargs):
        return self.obj(*args, **kwargs)

    @property
    def config_name(self):
        return self._config_name

    @property
    def services(self):
        return self._services

    @property
    def default_config(self):
        return self._default_config

    def check_config(self, config, name=''):
        # This is assuming that `config` is the config dictionary of
        # the *config parent*. That is, `config[self.config_name]`
        # exists.
        config = config.get(self.config_name, {})
        extras = set(config.keys()).difference(self.default_config)
        if len(extras) > 0:
            raise ConfigurationError(
                'Unsupported config options for "%s": %s'
                % (self.config_name, ', '.join(extras)))

        missing = set(self.default_config).difference(config)
        if len(extras) > 0:
            raise ConfigurationError(
                'Missing config options for "%s": %s'
                % (self.config_name, ', '.join(missing)))

        for other in self.services:
            if other in config:
                # I don't know what the right thing to do is here,
                # so be conservative and raise an error.
                #
                # N.B. I don't think this can happen when using auto-config
                # because Python will not let you have `arg` and `arg=val`
                # in the same parameter list. (`discover_config`, below,
                # guarantees that positional and named parameters are 
                # disjoint.)
                raise ProgrammerError(
                    'Configured object "%s" expects a '
                    '"%s" object to be available (from its '
                    'parameter list), but "%s" is already '
                    'defined as "%s" in its configuration.'
                    % (repr(self), other, other, config[other]))

    def _discover_config(self):
        '''
        Given an object at ``self.obj``, which must be a function,
        method or class, return a configuration *discovered* from
        the name of the object and its parameter list. This function
        is responsible for doing runtime reflection and providing
        understandable failure modes.

        The return value is a dictionary with three keys: ``name``,
        ``required`` and ``defaults``. ``name`` is the name of the
        function/method/class. ``required`` is a list of parameters
        *without* default values. ``defaults`` is a dictionary mapping
        parameter names to default values. The sets of parameter names in
        ``required`` and ``defaults`` are disjoint.

        When given a class, the parameters are taken from its ``__init__``
        method.

        Note that this function is purposefully conservative in the things
        that is will auto-configure. All of the following things will result
        in a :exc:`yakonfig.ProgrammerError` exception being raised:

        1. A parameter list that contains tuple unpacking. (This is invalid
           syntax in Python 3.)
        2. A parameter list that contains variable arguments (``*args``) or
           variable keyword words (``**kwargs``). This restriction forces
           an auto-configurable to explicitly state all configuration.

        Similarly, if given an object that isn't a function/method/class, a
        :exc:`yakonfig.ProgrammerError` will be raised.

        If reflection cannot be performed on ``obj``, then a ``TypeError``
        is raised.
        '''
        obj = self.obj
        skip_params = 0
        if inspect.isfunction(obj):
            name = obj.__name__
            inspect_obj = obj
            skip_params = 0
        elif inspect.ismethod(obj):
            name = obj.im_func.__name__
            inspect_obj = obj
            skip_params = 1  # self
        elif inspect.isclass(obj):
            if not hasattr(obj, '__init__'):
                raise ProgrammerError(
                    'Class "%s" does not have an "__init__" '
                    'method, so it cannot be auto configured.' % str(obj))
            name = obj.__name__
            if hasattr(obj, 'config_name'):
                name = obj.config_name
            inspect_obj = obj.__init__
            if not inspect.ismethod(inspect_obj):
                raise ProgrammerError(
                    '"%s.__init__" is not a method '
                    '(it is a "%s").' % (str(obj), type(obj)))
            skip_params = 1  # self
        else:
            raise ProgrammerError(
                'Expected a function, method or class to '
                'automatically configure, but got a "%s" '
                '(type: "%s").' % (repr(obj), type(obj)))

        argspec = inspect.getargspec(inspect_obj)
        if argspec.varargs is not None or argspec.keywords is not None:
            raise ProgrammerError(
                'The auto-configurable "%s" cannot contain '
                '"*args" or "**kwargs" in its list of '
                'parameters.' % repr(obj))
        if not all(isinstance(arg, strtype) for arg in argspec.args):
            raise ProgrammerError(
                'Expected an auto-configurable with no nested '
                'parameters, but "%s" seems to contain some '
                'tuple unpacking: "%s"'
                % (repr(obj), argspec.args))

        defaults = argspec.defaults or []
        # The index into `argspec.args` at which keyword arguments with default
        # values starts.
        i_defaults = len(argspec.args) - len(defaults)
        return {
            'name': name,
            'required': argspec.args[skip_params:i_defaults],
            'defaults': {k: defaults[i]
                         for i, k in enumerate(argspec.args[i_defaults:])},
        }



#!/usr/bin/env python

import fnmatch
import os
import sys
import subprocess

from setuptools import setup, find_packages
from distutils.core import Command

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'yakonfig'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
URL = 'http://github.com/diffeo/yakonfig'
DESC = 'load a configuration dictionary for a large application'

def read_file(file_name):
    file_path = os.path.join(
        os.path.dirname(__file__),
        file_name
        )
    return open(file_path).read()

def _myinstall(pkgspec):
    setup(
        script_args = ['-q', 'easy_install', '-v', pkgspec],
        script_name = 'easy_install'
    )

class PyTest(Command):
    '''run py.test'''
    description = 'runs py.test to execute all tests'
    user_options = []
    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        if self.distribution.install_requires:
            for ir in self.distribution.install_requires:
                _myinstall(ir)
        if self.distribution.tests_require:
            for ir in self.distribution.tests_require:
                _myinstall(ir)

        # reload sys.path for any new libraries installed
        import site
        site.main()
        print sys.path
        pytest = __import__('pytest')
        if pytest.main(['-n', '8', '-vvs', 'yakonfig']):
            sys.exit(1)

setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    license=read_file('LICENSE.txt'),
    long_description=read_file('README.md'),
    #source_label=SOURCE_LABEL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages = find_packages(),
    cmdclass={'test': PyTest,},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',  ## MIT/X11 license http://opensource.org/licenses/MIT
    ],
    install_requires=[
        'pyyaml',
    ],
    tests_require=[
        'pytest',
        'ipdb',
        'pytest-cov',
        'pytest-xdist',
        'pytest-timeout',
        'pytest-incremental',
        'pytest-capturelog',
    ],
)



'''Global configuration management tools.

.. This software is released under an MIT/X11 open source license.
   Copyright 2013-2014 Diffeo, Inc.

Loads a YAML file (with extensions) and makes it available globally to
a Python application.  Application code can call
:func:`yakonfig.get_global_config` to get values from the
configuration store.

In typical use, a program will declare some number of top-level
configuration modules.  These modules mimic the structure of
:class:`yakonfig.Configurable`, and are passed in directly to
:func:`yakonfig.parse_args()`.  This makes the corresponding
configuration available in the global configuration.  For example::

    # I am a_module.py
    import yakonfig

    # Configurable metadata
    config_name = 'a_module'
    default_config = { 'message': 'hello world' }
    def add_arguments(parser):
        parser.add_argument('--message')
    runtime_keys = { 'message': 'message' }

    def main():
        parser = argparse.ArgumentParser()
        args = yakonfig.parse_args(parser, [yakonfig, a_module])
        print yakonfig.get_global_config('a_module', 'message')
    if __name__ == '__main__':
        main()

Running ``python -m a_module`` will print out "hello world"; running
``python -m a_module --message goodbye`` will inject the "goodbye"
message into the configuration, and then print it out.

.. note:: Avoid using :meth:`argparse.ArgumentParser.set_defaults` in
          combination with :func:`yakonfig.parse_args`.  Any default
          values set this way will override values set in the user's
          configuration file.

Most of the useful functions and objects in the submodules are
re-exported from the top-level module.

Common command-line arguments
=============================

All programs based on :mod:`yakonfig` support the following
command-line arguments.

.. program:: yakonfig

.. option:: --config <file.yaml>, -c <file.yaml>

Load ``file.yaml`` as the primary configuration file.  Options set in
the configuration file override any default values; options set on the
command line override values in this configuration file.

.. option:: --dump-config {default|effective|full}

Write the configuration as YAML to standard output, then stop
immediately.  The argument indicates which configuration is dumped: if
``default`` then the default configuration, ignoring all user setup,
is dumped; if ``full`` then the dump contains the complete
configuration, including all default settings, settings from the
configuration file, and command-line options; and if ``effective``
then the dump contains only those settings that differ from the
defaults, producing the minimum configuration to recreate the current
settings.

YAML extensions
===============

``key: !include path``
  Loads a yaml file at `path` and inserts it as the value associated
  with `key`.

``key: !include_yaml path`` *(deprecated)*
  Alias for ``!include``.

``key: !include_func package.path.to.func`` *(deprecated)*
  Calls a python function. from a fully specified name of
  ``package.func`` If the function name ends in "yaml" the return
  value is interpreted as a yaml document body in a string and
  parsed. Otherwise the return value is assumed to be a dict or other
  object that can be simply assigned to they key at this point in the
  yaml file.

Top-level entry points
======================

.. autofunction:: parse_args
.. autofunction:: set_default_config
.. autofunction:: defaulted_config
.. autofunction:: get_global_config
.. autofunction:: set_global_config
.. autofunction:: clear_global_config

Configurable modules
====================

Modules that can be configured by the user implement the
:class:`yakonfig.Configurable` interface, or more often, have their
classes and modules provide the same names.  Anything that includes a
:attr:`~yakonfig.Configurable.config_name` property can be passed into
the top-level entry points above.

.. autoclass:: Configurable
   :members:

.. autoclass:: ProxyConfigurable
   :members:

.. autoclass:: NewSubModules
   :members:

.. autofunction:: check_toplevel_config
.. autofunction:: check_subconfig

Operations on configuration
===========================

Configuration is always passed around as plain Python dictionaries.
If you can guarantee that a configuration has passed through yakonfig,
then you can generally guarantee that its default values are present
and that any defined checks have passed.

.. autofunction:: overlay_config
.. autofunction:: diff_config

Exceptions
==========

.. autoclass:: ConfigurationError
.. autoclass:: ProgrammerError

Interactive programs
====================

.. currentmodule:: yakonfig.cmd
.. autoclass:: ArgParseCmd
.. currentmodule:: yakonfig

'''
from __future__ import absolute_import
from yakonfig.configurable import Configurable, ProxyConfigurable, \
    NewSubModules, check_subconfig
from yakonfig.exceptions import *
from yakonfig.merge import diff_config, overlay_config
from yakonfig.toplevel import parse_args, set_default_config, \
    defaulted_config, check_toplevel_config, \
    config_name, add_arguments, runtime_keys
from yakonfig.yakonfig import clear_global_config, set_global_config, \
    get_global_config
